package com.becan.pi.audio;

import java.io.InputStream;
import java.net.Socket;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.SourceDataLine;

public class Player {

	private int PREBUFFER;
	private int BUFFER_SIZE;
	private int AUDIO_LINE_BUFFER_SIZE;
	private int PACKET_SIZE;

	private boolean isRunning = true;

	private Socket connection;
	private InputStream in;

	private Buffer buffer;

	private SourceDataLine audioLine;

	public Player(Socket connection, int prebuffer) {
		this.connection = connection;
		try {
			this.in = connection.getInputStream();
		} catch (Exception e) {
			e.printStackTrace();
		}
		this.PREBUFFER = prebuffer;
	}

	public void start() {
		openAudioLine();
		int frameSize = audioLine.getFormat().getFrameSize();
		int bytesPerMs = (48000 * frameSize) / 1000;
		BUFFER_SIZE = PREBUFFER * bytesPerMs * 3;
		PACKET_SIZE = (int) (AUDIO_LINE_BUFFER_SIZE * 0.8);
		PACKET_SIZE = PACKET_SIZE - PACKET_SIZE % frameSize;
		buffer = new Buffer(BUFFER_SIZE, PACKET_SIZE);

		Thread read = new Thread(reader);
		read.setPriority(Thread.MIN_PRIORITY);
		read.start();

		Thread write = new Thread(writer);
		write.setPriority(Thread.MAX_PRIORITY);
		write.start();
	}

	private void close() {
		try {
			isRunning = false;
			connection.close();
			audioLine.removeLineListener(lineListener);
			audioLine.flush();
			audioLine.stop();
			audioLine.close();
			buffer = null;
		} catch (Exception e) {
		}
	}

	private void openAudioLine() {
		try {
			AudioFormat af = new AudioFormat(48000, 16, 2, true, false);
			DataLine.Info dataLineInfo = new DataLine.Info(SourceDataLine.class, af, 65536);
			audioLine = (SourceDataLine) AudioSystem.getLine(dataLineInfo);
			audioLine.open(af, 65536);
			audioLine.start();
			AUDIO_LINE_BUFFER_SIZE = audioLine.getBufferSize();
			audioLine.addLineListener(lineListener);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private LineListener lineListener = new LineListener() {
		@Override
		public void update(LineEvent e) {
			if (e.getType() == LineEvent.Type.STOP || e.getType() == LineEvent.Type.CLOSE) {
				close();
			}
		}
	};

	private Runnable reader = new Runnable() {
		@Override
		public void run() {
			while (isRunning) {
				read();
			}
		}

		public void read() {
			try {
				byte[] chunk = buffer.popFree();

				int count = 0;
				do {
					int c = in.read(chunk, count, chunk.length - count);
					if (c == -1) {
						close();
						return;
					}
					count += c;
				} while (count != chunk.length);

				buffer.push(chunk);
			} catch (Exception e) {
				close();
				e.printStackTrace();
			}
		}
	};

	private Runnable writer = new Runnable() {
		@Override
		public void run() {
			try {
				Thread.sleep(PREBUFFER);
			} catch (Exception e) {
				e.printStackTrace();
			}

			while (isRunning) {
				write();
			}
		}

		public void write() {
			try {
				byte[] chunk = buffer.pop();
				audioLine.write(chunk, 0, chunk.length);
				buffer.pushFree(chunk);
			} catch (Exception e) {
				close();
				e.printStackTrace();
			}
		}
	};

}
